#!/usr/bin/python2.2

## redhat-config-xfree - an XFree86 configuration tool
## Copyright (c) 2002 Red Hat, Inc.

## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import string
import gobject
import sys
import time
import os
import signal
import rhpl.monitor
import rhpl.videocard
import rhpl.mouse
import rhpl.keyboard
import xf86config
import math
import getopt
import tempfile
import locale
from rhpl.translate import _, N_,textdomain

TRUE = 1
FALSE = 0

# TODO:
# probe hsync/vsync seems to give different results than db, which one to trust?
# Better way to detect when DRI is availible.

all_depths = (8,15,16,24)

logFile = "/var/log/XFree86.setup.log"

def verbose(str):
    if verbose_output:
        print str

def resolution_from_string(res):
    (w, h) = string.split (res, "x")
    return (string.atoi(w), string.atoi(h))

def resolution_area(res):
    (w, h) = resolution_from_string(res)
    return w * h

def compare_resolution(res1, res2):
    a1 = resolution_area(res1)
    a2 = resolution_area(res2)
    if a1 < a2:
        return -1
    if a1 == a2:
        return 0
    return 1

def ranges_to_string(array, len):
    str = ""
    for i in range(len):
        r = array[i]
        if str != "":
            str = str + ","
        if r[0] == r[1]:
            str = str + repr(r[0])
        else:
            str = str + repr(r[0]) + "-"  + repr(r[1])
    return str

def string_to_ranges(str):
    l = []
    pieces = string.split(str, ",")
    for piece in pieces:
        tmp = string.split(piece, "-")
        if len(tmp) == 1:
            l.append( (string.atof(tmp[0]), string.atof(tmp[0])) )
        else:
            l.append( (string.atof(tmp[0]), string.atof(tmp[1])) )
    return l

def dpi_string(resolution, physical_width, physical_height):
    if physical_width == 0 or physical_height == 0:
        return _("Unknown")
    (w,h) = resolution_from_string(resolution)
    dpix = int(round(w * 25.4/physical_width))
    dpiy = int(round(h * 25.4/physical_height))
    return _("%d by %d dots per inch")%(dpix, dpiy)

def unique(s):
    if len(s) == 0:
        return []
    u = {}
    for x in s:
        u[x] = 1
    return u.keys()

def compare_resolution(res1, res2):
    r1 = resolution_from_string(res1)
    r2 = resolution_from_string(res2)
    if r1 < r2:
        return -1
    if r1 == r2:
        return 0
    return 1

def tokenize_line(line):
    list = []
    current = ""
    
    in_quote = 0
    for c in line:
        if in_quote:
            if c == '"':
                in_quote = 0
                list.append(current)
                current = ""
            else:
                current = current + c
        else:
            if c =='"':
                if len(current) > 0:
                    list.append(current)
                    current = ""
                in_quote = 1
            elif c == "#":
                break
            elif c in string.whitespace:
                if len(current) > 0:
                    list.append(current)
                    current = ""
            else:
                current = current + c
    if len(current) > 0:
        list.append(current)
    return list
            


class HardwareState:
    def __init__(self, xconfig):
        verbose(_("Probing monitor"))
        self.monitor = rhpl.monitor.MonitorInfo()
        verbose(_("Probing videocard"))
        self.videocard = rhpl.videocard.VideoCardInfo()

        # Init videocard
        self.colordepth = 16
        self.resolution = "800x600"
        self.video_ram = 0
        self.probed_video_ram = 0
        self.videocard_options = []
        self.dri_enabled = 0
        self.videocard_name = None
        
        self.monitor_name = None
        self.hsync = None
        self.vsync = None
        self.physical_width = 0
        self.physical_height = 0
        self.probed_hsync = None
        self.probed_vsync = None
        self.probed_physical_width = 0
        self.probed_physical_height = 0

        self.all_resolutions = ["640x480", "800x600", "1024x768", "1152x864", "1280x960", "1280x1024", "1400x1050", "1600x1200"]
        
        self.init_from_probed()

        if xconfig:
            self.init_from_xconfig(xconfig)

        # Handle completely unset things:
                
        if self.videocard_name == None:
            self.videocard_name = "Unknown video card"
        
        if self.videocard_driver == None:
            self.videocard_driver = "vga"

        # Default to: Generic Extended Super VGA H: 31.5-37.9 V: 50-70
        if self.hsync == None:
            self.hsync = "31.5-37.9"
        if self.vsync == None:
            self.vsync = "50-70"
        if self.monitor_name == None:
            self.monitor_name = "Unknown monitor"
        
    def parse_lines(self, lines):
        for line in string.split(lines, "\n"):
            items = tokenize_line(line)
            if len(items) > 1 and string.upper(items[0]) == "OPTION":
                name = items[1]
                if len(items) > 2:
                    val = items[2]
                else:
                    val = ""
                self.videocard_options.append( (name, val) )
                
    def init_from_probed(self):
        primary = None
        if self.videocard:
            primary = self.videocard.primaryCard()

        # Set up videocard name
        if primary != None:
            devid = primary.getDevID()
            if devid != None:
                self.videocard_name = devid
        self.videocard_driver = None

        if primary:
            mem = primary.getVideoRam()
            if mem:
                self.probed_video_ram = string.atoi(mem)

        # Set up driver
        if primary:
            carddata = primary.getCardData()
            if carddata:
                if carddata.has_key("DRIVER"):
                    self.videocard_driver = carddata["DRIVER"]

                    # Add options if any, these are overridden
                    # if there are other options in the x config
                    if carddata.has_key("LINE"):
                        lines = carddata["LINE"]
                        self.parse_lines(lines)


        if self.monitor:
            name = self.monitor.getMonitorName()
            if name != None and (self.monitor_name == None or self.monitor_name == "Monitor Model"):
                self.monitor_name = name

            self.probed_hsync = self.monitor.getMonitorHorizSync(TRUE)
            self.probed_vsync = self.monitor.getMonitorVertSync(TRUE)

            self.probed_physical_width = self.monitor.getMonitorPhysicalWidth()
            self.probed_physical_height = self.monitor.getMonitorPhysicalHeight()

            self.hsync = self.probed_hsync
            self.vsync = self.probed_vsync
            self.physical_width = self.probed_physical_width
            self.physical_height = self.probed_physical_height
        
    def init_from_xconfig(self, xconfig):
        screen = xf86config.getPrimaryScreen(xconfig)
        device = xf86config.lookupDevice(xconfig, screen.device)
        monitor = xf86config.lookupMonitor(xconfig, screen.monitor)
            
        if screen:
            self.colordepth = screen.defaultdepth
            for d in screen.display:
                if d.depth == self.colordepth:
                    self.resolution = d.modes[0].name
                    break

        if screen:
            for d in screen.display:
                for mode in d.modes:
                    res = mode.name
                    if res  != None and res not in self.all_resolutions:
                        self.all_resolutions.append(res)
            self.all_resolutions.sort (compare_resolution)

        if device:
            self.videocard_name = device.board

        if device:
            self.video_ram = device.videoram

        if device != None:
            self.videocard_driver = device.driver

        # Set up driver options
        if device != None:
            if len(device.options):
                # Override the probed values
                self.videocard_options = []
            for option in device.options:
                self.videocard_options.append( (option.name, option.val) )

        if monitor != None:
            self.monitor_name = monitor.modelname
            if monitor.n_hsync > 0:
                self.hsync = ranges_to_string(monitor.hsync, monitor.n_hsync)
            if monitor.n_vrefresh > 0:
                self.vsync = ranges_to_string(monitor.vrefresh, monitor.n_vrefresh)
            self.physical_width = monitor.width
            self.physical_height = monitor.height
            
        # Is DRI enabled?
        for l in xconfig.modules.load:
            if l.name == "dri":
                self.dri_enabled = 1
                break
   
    def merge_into(self, xconfig):
        screen = xf86config.getPrimaryScreen(xconfig)

        #TODO: Handle no screen?

        screen.defaultdepth = self.colordepth

        display = None
        for d in screen.display:
            if d.depth == self.colordepth:
                display = d
                break

        # Do we need to add a subsection for the current depth?
        if display == None:
            display = xf86config.XF86ConfDisplay()
            display.depth = self.colordepth
            screen.display.insert(display)

        while len(display.modes) > 0:
            display.modes.remove(0)
        display.modes.insert(xf86config.XF86Mode(self.resolution))
        # Add all other availible resolutions too.
        availible = self.availible_resolutions()
        r = -1
        
        for i in range(0, len(availible)):
            if compare_resolution (availible[i], self.resolution) >=  0:
                r = i - 1
                break

        for i in range (r, -1, -1):
            display.modes.insert(xf86config.XF86Mode(availible[i]))

        device = xf86config.lookupDevice(xconfig, screen.device)
        #TODO: Handle no device

        device.board = self.videocard_name
        device.videoram = self.video_ram
        device.driver = self.videocard_driver
        while len(device.options) > 0:
            device.options.remove(0)
            
        for option in self.videocard_options:
            device.options.insert(xf86config.XF86Option(option[0], option[1]))

        monitor = xf86config.lookupMonitor(xconfig, screen.monitor)
        #TODO: Handle no monitor

        monitor.modelname = self.monitor_name
        
        hsync = string_to_ranges(self.hsync)
        monitor.n_hsync = len(hsync)
        for i in range(len(hsync)):
            monitor.hsync[i] = hsync[i]

        vsync = string_to_ranges(self.vsync)
        monitor.n_vrefresh = len(vsync)
        for i in range(len(vsync)):
            monitor.vrefresh[i] = vsync[i]

        monitor.width = int(round(self.physical_width))
        monitor.height = int(round(self.physical_height))

        dri_index = -1
        for i in range(len(xconfig.modules.load)):
            if xconfig.modules.load[i].name == "dri":
                dri_index = i
                break
        
        if self.dri_enabled:
            if dri_index < 0:
                xconfig.modules.load.insert(xf86config.XF86ConfLoad("dri"))
        else:
            if dri_index >= 0:
                xconfig.modules.load.remove(dri_index)
        
    def recalc_mode(self):
        availible = self.availible_resolutions()
        if not self.resolution in availible:
            self.resolution = availible[-1]
        availible = self.availible_color_depths()
        if not self.colordepth in availible:
            self.colordepth = availible[-1]

    def set_resolution(self, resolution):
        self.resolution = resolution
        self.recalc_mode()

    def set_colordepth(self, depth):
        self.colordepth = depth
        self.recalc_mode()

    def set_dri_enabled(self, enabled):
        self.dri_enabled = enabled

    def set_monitor_name(self, name):
        self.monitor_name = name

    def set_hsync(self, hsync):
        self.hsync = hsync

    def set_vsync(self, vsync):
        self.vsync = vsync

    def set_physical_size(self, w, h):
        self.physical_width = w
        self.physical_height = h
    
    def set_videocard_name(self, name):
        self.videocard_name = name
    
    def set_videocard_driver(self, driver):
        self.videocard_driver = driver
    
    def set_videocard_card(self, card):
	try:
	    foo = rhpl.videocard.Video_cardsDBLookup(card)
	    if foo:
		if foo.has_key('DRIVER'):
		    self.videocard_driver = foo['DRIVER']
		if foo.has_key("LINE"):
		    lines = carddata["LINE"]
		    self.parse_lines(lines)
	except:
	    pass
	
    def set_videocard_ram(self, ram):
        self.video_ram = ram
    
    def set_videocard_options(self, options):
        self.videocard_options = options
    
    def get_resolution(self):
        return self.resolution

    def get_colordepth(self):
        return self.colordepth

    def get_videocard_name(self):
        return self.videocard_name
    
    def get_videocard_driver(self):
        return self.videocard_driver
    
    def get_videocard_ram(self):
        return self.video_ram
    
    def get_videocard_ram_or_probed(self):
        if self.video_ram == 0:
            return self.probed_video_ram
        return self.video_ram

    def get_all_resolutions(self):
        return self.all_resolutions
    
    def get_hsync(self):
        return self.hsync
    
    def get_hsync_or_probed(self):
        if self.hsync == None:
            return self.probed_hsync
        return self.hsync
    
    def get_vsync(self):
        return self.vsync
    
    def get_vsync_or_probed(self):
        if self.vsync == None:
            return self.probed_vsync
        return self.vsync
    
    def get_physical_size(self):
        return (self.physical_width, self.physical_height)

    def get_physical_size_or_probed(self):
        if self.physical_width == 0 or self.physical_height == 0:
            return (self.probed_physical_width, self.probed_physical_height)
        return (self.physical_width, self.physical_height)

    def get_monitor_name(self):
    	return self.monitor_name

    def get_dri_enabled(self):
    	return self.dri_enabled

    def get_videocard_options(self):
        return self.videocard_options
                
    def availableModes(self):
        modes = { "8" : [ "640x480", "800x600" ] }

        ram = self.get_videocard_ram_or_probed()
        # If we can't probe, assume at least 8 megs.
        if ram == 0:
            ram = 8*1024

        modes["8"] = filter (lambda res: resolution_area(res)*1 <= ram*1024, self.all_resolutions)
        modes["16"] = filter (lambda res: resolution_area(res)*2 <= ram*1024, self.all_resolutions)
        modes["24"] = filter (lambda res: resolution_area(res)*4 <= ram*1024, self.all_resolutions)

        return modes

    def availible_resolutions(self):
        global monitor
        modes = self.availableModes()
        l = []
        if modes.has_key("8"):
            l = l + modes["8"]
        if modes.has_key("16"):
            l = l + modes["16"]
        if modes.has_key("24"):
            l = l + modes["24"]

        availible = []
        for res in unique(l):
            # This means non-vesa modes are always "supported". Hard to do otherwise.
            if rhpl.monitor.monitor_supports_mode(self.hsync, self.vsync, res):
                availible.append(res)

        availible.sort(compare_resolution)

        # Need a fallback
        if len(availible) == 0:
            availible = ["640x480"]
            
        return availible

    def availible_color_depths(self):
        modes = self.availableModes()
        l = []
        if modes.has_key("8") and self.resolution in modes["8"]:
            l.append(8)
        if modes.has_key("16") and self.resolution in modes["16"]:
            l.append(15)
            l.append(16)
        if modes.has_key("24") and self.resolution in modes["24"]:
            l.append(24)
        l.sort()
        return l

    def supports_dri(self):
        #TODO: This check is lame:
        driver = self.get_videocard_driver()
        if driver in ["mga", "radeon"]:
            return 1
        return 0
    

class ScreenSizePreview:
    max_width = 1600
    max_height = 1200
        
    def screen_expose(self, widget, event):
        gc = widget.window.new_gc()
	gc.set_rgb_fg_color(widget.style.black)
        widget.window.draw_rectangle(gc, TRUE, 0,0, self.widget_width, self.widget_height)
        self.screenshot.render_to_drawable_alpha(widget.window,
                                                  0, 0,
                                                  (self.widget_width - self.screenshot.get_width())/2,
                                                  (self.widget_height - self.screenshot.get_height())/2,
                                                  self.screenshot.get_width(), self.screenshot.get_height(),
                                                  gtk.gdk.PIXBUF_ALPHA_FULL,
                                                  0, gtk.gdk.RGB_DITHER_MAX, 0, 0)
        w = self.res_x * self.widget_width / self.max_width - 1
        h = self.res_y * self.widget_height / self.max_height - 1
        widget.window.draw_rectangle(self.gc, FALSE,
                                     (self.widget_width - w)/2,
                                     (self.widget_height - h)/2,
                                     w,h)

    def realize(self, widget):
        self.gc = widget.window.new_gc()
        self.gc.set_background(widget.get_style().white)
        self.gc.set_foreground(widget.get_style().black)
        self.gc.set_line_attributes(0, gtk.gdk.LINE_DOUBLE_DASH,
                                    gtk.gdk.CAP_BUTT, gtk.gdk.JOIN_MITER)

    def __init__(self, widget, widget_width, widget_height):
        self.widget_width = widget_width
        self.widget_height = widget_height
        self.widget = None
        
        screenshot = gtk.gdk.Pixbuf(gtk.gdk.COLORSPACE_RGB, FALSE, 8,
                                    gtk.gdk.screen_width(), gtk.gdk.screen_height())
        screenshot.get_from_drawable(gtk.gdk.get_default_root_window(),
                                     gtk.gdk.colormap_get_system(),
                                     0, 0, 0, 0,
                                     gtk.gdk.screen_width(), gtk.gdk.screen_height())
        self.screenshot = screenshot.scale_simple(gtk.gdk.screen_width() * widget_width / self.max_width,
                                                  gtk.gdk.screen_height() * widget_height / self.max_height,
                                                  gtk.gdk.INTERP_BILINEAR)

        self.connect_to(widget)


    def set_resolution(self, res_x, res_y):
        self.res_x = res_x
        self.res_y = res_y
        if self.widget != None:
            self.widget.queue_draw()

    def connect_to(self, widget):
        self.widget = widget
        widget.connect("expose_event", self.screen_expose)
        widget.connect_after("realize", self.realize)


class MonitorDialog:
    def selection_changed(self, selection):
        if self.changing_selection:
            return
        (store, iter) = selection.get_selected()
        if iter == None:
            return

        if store.get_value(iter, 1) != None:
            model = store.get_value(iter, 0)
            vsync = store.get_value(iter, 2)
            hsync = store.get_value(iter, 3)
            self.model_entry.set_text(model)
            self.hsync_entry.set_text(hsync)
            self.vsync_entry.set_text(vsync)

    def probe_monitor(self, widget):
        if self.probed_path:
            self.goto_path(self.probed_path)
        model = self.monitor.getMonitorName()
        hsync = self.monitor.getMonitorHorizSync(TRUE)
        vsync = self.monitor.getMonitorVertSync(TRUE)
        self.model_entry.set_text(model)
        self.hsync_entry.set_text(hsync)
        self.vsync_entry.set_text(vsync)
            
    def __init__(self, xml, monitor, probed_id):
        self.xml = xml
        self.monitor = monitor

        self.model_entry = self.xml.get_widget("monitor_name_entry")
        self.hsync_entry = self.xml.get_widget("monitor_hsync_entry")
        self.vsync_entry = self.xml.get_widget("monitor_vsync_entry")

        self.probed_path = None

        self.changing_selection = 0
        
        dialog = self.xml.get_widget("monitor_dialog")
        dialog.set_transient_for(self.xml.get_widget("display_dialog"))
        dialog.set_modal(TRUE)

        monitor_store = gtk.TreeStore(gobject.TYPE_STRING, gobject.TYPE_STRING,
                                      gobject.TYPE_STRING, gobject.TYPE_STRING)
        self.monitor_store = monitor_store

        self.models = {}
        db = monitor.readMonitorsDB()
        l = db.keys()
        l.sort()
        for maker in ["Generic"] + l:
            parent = monitor_store.append(None)
            monitor_store.set_value(parent, 0, maker)
            for mon in db[maker]:
                iter = monitor_store.append(parent)
                monitor_store.set_value(iter, 0, mon[0])
                monitor_store.set_value(iter, 1, mon[1])
                monitor_store.set_value(iter, 2, mon[2])
                monitor_store.set_value(iter, 3, mon[3])
                
                if probed_id and string.upper(mon[1]) == string.upper(probed_id):
                    self.probed_path = monitor_store.get_path(iter) 

                self.models[string.upper(mon[0])] = monitor_store.get_path(iter) 

        tree_view = self.xml.get_widget("monitor_tree")
        tree_view.set_model(monitor_store)
        
        selection = tree_view.get_selection()
        selection.connect("changed", self.selection_changed)

        self.col = gtk.TreeViewColumn(None, gtk.CellRendererText(), text=0)
        tree_view.append_column(self.col)

        button = self.xml.get_widget("monitor_probe_button")
        button.connect("clicked", self.probe_monitor)
        button.set_sensitive(probed_id != "Unprobed Monitor")

    def goto_path(self, path):
        tree_view = self.xml.get_widget("monitor_tree")
        tree_view.expand_row(path[:-1], TRUE)
        self.changing_selection = 1
        selection = tree_view.get_selection()
        selection.unselect_all()
        iter = self.monitor_store.get_iter(path)
        selection.select_iter(iter)
        self.changing_selection = 0
        tree_view.set_cursor(path, self.col, FALSE)
        tree_view.scroll_to_cell(path, self.col, TRUE, 0.5, 0.0)
        tree_view.grab_focus ()

    def hydrate(self, state):
        self.model_entry.set_text(state.get_monitor_name())
        self.hsync_entry.set_text(state.get_hsync_or_probed())
        self.vsync_entry.set_text(state.get_vsync_or_probed())

    def dehydrate(self, state):
        model = self.model_entry.get_text()
        state.set_monitor_name(model)
        hsync = self.hsync_entry.get_text()
        state.set_hsync(hsync)
        vsync = self.vsync_entry.get_text()
        state.set_vsync(vsync)

    def run(self):
        model = self.model_entry.get_text()
        key = string.upper(model)
        if self.models.has_key(key):
            self.goto_path(self.models[key])
        
        dialog = self.xml.get_widget("monitor_dialog")
        dialog.present()
        while 1:
            res = dialog.run()
            if res != gtk.RESPONSE_OK:
                dialog.hide()
                return FALSE
            
            hsync = self.hsync_entry.get_text()
            if not rhpl.monitor.isValidSyncRange(hsync):
                m = gtk.MessageDialog(dialog, gtk.DIALOG_MODAL, gtk.MESSAGE_ERROR, gtk.BUTTONS_OK, _("Invalid hsync specification"))
                m.run()
                m.destroy()
                continue
            
            vsync = self.vsync_entry.get_text()
            if not rhpl.monitor.isValidSyncRange(vsync):
                m = gtk.MessageDialog(dialog, gtk.DIALOG_MODAL, gtk.MESSAGE_ERROR, gtk.BUTTONS_OK, _("Invalid vsync specification"))
                m.run()
                m.destroy()
                continue

            break
        
        dialog.hide()
        return TRUE

class DPIDialog:
    unit_inch = 0
    unit_millimeter = 1
    
    def __init__(self, xml, monitor):
        self.xml = xml
        self.monitor = monitor

        self.unit = self.unit_millimeter
        self.resolution = None
        
        dialog = xml.get_widget("dpi_dialog")
        dialog.set_transient_for(xml.get_widget("display_dialog"))
        dialog.set_modal(TRUE)
        
        self.width_spin = xml.get_widget("dpi_width_spin")
        self.height_spin = xml.get_widget("dpi_height_spin")
        self.unit_menu = xml.get_widget("dpi_unit_menu")
        self.dpi_label = xml.get_widget("dpi_resolution_label")

        # Initial unit is millimeter
        self.width_spin.set_digits(0)
        self.height_spin.set_digits(0)
            
        self.unit_menu.connect("changed", self.units_changed)
        self.width_spin.connect("value_changed", self.value_changed)
        self.height_spin.connect("value_changed", self.value_changed)

        button = xml.get_widget("dpi_probe_button")
        button.connect("clicked", self.probe_monitor)
        if monitor.getMonitorPhysicalWidth(TRUE) == 0 or monitor.getMonitorPhysicalHeight(TRUE) == 0:
            button.set_sensitive(FALSE)

    def probe_monitor(self, widget):
        self.set_value(self.monitor.getMonitorPhysicalWidth(TRUE),
                       self.monitor.getMonitorPhysicalHeight(TRUE))

    def units_changed(self, widget):
        (w, h) = self.get_value()
        self.unit = widget.get_history()
        self.set_value(w,h)
        if self.unit == self.unit_inch:
            self.width_spin.set_digits(2)
            self.height_spin.set_digits(2)
        else:
            self.width_spin.set_digits(0)
            self.height_spin.set_digits(0)
            

    def update_dpi_label(self):
        (w,h) = self.get_value()
        if self.resolution:
            self.dpi_label.set_text(_("Current resolution: ") + dpi_string(self.resolution, w, h))

    def value_changed(self, widget):
        self.update_dpi_label()
        
    def set_value(self, w, h):
        if self.unit == self.unit_inch:
            self.width_spin.set_value(w / 25.4)
            self.height_spin.set_value(h / 25.4)
        else:
            self.width_spin.set_value(w)
            self.height_spin.set_value(h)
        
    def get_value(self):
        if self.unit == self.unit_inch:
            w = self.width_spin.get_value() * 25.4
            h = self.height_spin.get_value() * 25.4
        else:
            w = self.width_spin.get_value()
            h = self.height_spin.get_value()
        w = math.floor(w+0.5)
        h = math.floor(h+0.5)
        return (w, h)
        
    def hydrate(self, state):
        (w, h) = state.get_physical_size_or_probed()
        self.set_value (w, h)
        self.resolution = state.get_resolution()
        self.update_dpi_label()
    
    def dehydrate(self, state):
        (realw, realh) = state.get_physical_size()
        probed = state.get_physical_size_or_probed()
        inputed = self.get_value()
        # If the size is currently unspecified, and the user
        # didnt change from the probed size, keep it unspecified
        if (realw == 0 or realh == 0) and probed == inputed:
            return
        state.set_physical_size(inputed[0], inputed[1])

    def run(self):
        dialog = self.xml.get_widget("dpi_dialog")
        dialog.present()
        while 1:
            res = dialog.run()
            if res != gtk.RESPONSE_OK:
                dialog.hide()
                return FALSE

            break
        
        dialog.hide()
        return TRUE


class VideocardDialog:
    #TODO: Is this list correct?
    drivers = ["mga", "ati", "chips", "cirrus", "cyrix", "fbdev",
               "glint", "i128", "i740", "i810", "mga", "neomagic",
               "newport", "nv", "r128", "radeon", "rendition", "s3",
               "s3virge", "savage", "siliconmotion", "sis", "tdfx",
               "trident", "tseng", "vesa", "vga", "vmware"]

    def __init__(self, xml, videocard):
        self.xml = xml
        self.videocard = videocard

        self.changing_selection = 0
        
        dialog = xml.get_widget("videocard_dialog")
        dialog.set_transient_for(xml.get_widget("display_dialog"))
        dialog.set_modal(TRUE)

        optionmenu = xml.get_widget("videocard_mem_option_menu")
        menu = gtk.Menu()
        for ram in videocard.possible_ram_sizes():
            if ram % 1024 != 0:
                m = (ram, repr(ram) +" Kb")
            else:
                m = (ram, repr(ram/1024) +" Mb")
            item = gtk.MenuItem(m[1])
            item.set_data("mem-size", m[0])
            item.show()
            menu.append(item)
        optionmenu.set_menu(menu)

        xml.get_widget("videocard_driver_combo"). set_popdown_strings(self.drivers)

        videocard_store = gtk.ListStore(gobject.TYPE_STRING)
        self.videocard_store = videocard_store

        iter = videocard_store.append()
        videocard_store.set_value(iter, 0, "Custom")
        card_list = videocard.cardsDB().keys()
        card_list.sort()
        for name in card_list:
            iter = videocard_store.append()
            videocard_store.set_value(iter, 0, name)

        tree_view = xml.get_widget("videocard_tree")
        tree_view.set_model(videocard_store)
        selection = tree_view.get_selection()
        # Select the first before setting up callback, to avoid first callback.
        selection.select_iter(videocard_store.get_iter_first())
        selection.connect("changed", self.selection_changed)

        self.vc_col = gtk.TreeViewColumn(None, gtk.CellRendererText(), text=0)
        tree_view.append_column(self.vc_col)

        options_store = gtk.ListStore(gobject.TYPE_STRING, gobject.TYPE_STRING)
        self.options_store = options_store
        tree_view = xml.get_widget("videocard_options_tree")
        tree_view.set_model(options_store)
        cell_renderer = gtk.CellRendererText()
        cell_renderer.set_property("editable", TRUE)
        cell_renderer.connect("edited", self.option_name_edited)
        self.col = gtk.TreeViewColumn("option", cell_renderer, text=0)
        tree_view.append_column(self.col)
        cell_renderer = gtk.CellRendererText()
        cell_renderer.set_property("editable", TRUE)
        cell_renderer.connect("edited", self.option_value_edited)
        self.col = gtk.TreeViewColumn("value", cell_renderer, text=1)
        tree_view.append_column(self.col)
        self.options_tree_view = tree_view

        self.type_entry = xml.get_widget("videocard_type_entry")
        self.driver_entry = xml.get_widget("videocard_driver_entry")
        self.mem_check = xml.get_widget("videocard_mem_check")
        self.mem_optionmenu = xml.get_widget("videocard_mem_option_menu")

        self.mem_check.connect("toggled", self.custom_mem_toggled)

        button = xml.get_widget("videocard_probe_button")
        button.connect("clicked", self.probe_videocard)
        primary = videocard.primaryCard()
        button.set_sensitive(primary != None and
                             primary.getCardData() != None)

        button = xml.get_widget("videocard_add_option_button")
        button.connect("clicked", self.add_option)
        button = xml.get_widget("videocard_remove_option_button")
        button.connect("clicked", self.remove_option)

    def parse_lines(self, lines):
        for line in string.split(lines, "\n"):
            items = tokenize_line(line)
            if len(items) > 1 and string.upper(items[0]) == "OPTION":
                iter = self.options_store.append()
                self.options_store.set_value(iter, 0, items[1])
                if len(items) > 2:
                    self.options_store.set_value(iter, 1, items[2])
                else:
                    self.options_store.set_value(iter, 1, "")
            if len(items) == 2 and string.upper(items[0]) == "VIDEORAM":
                self.set_custom_memory(TRUE)
                self.set_memory_size(string.atoi(items[1]))

    def goto_row(self, cardname):
        iter = self.videocard_store.get_iter_first()
        while iter:
            if self.videocard_store.get_value(iter, 0) == cardname:
                tree_view = self.xml.get_widget("videocard_tree")
                self.changing_selection = 1
                selection = tree_view.get_selection()
                selection.select_iter(iter)
                self.changing_selection = 0
                path = self.videocard_store.get_path(iter)
                tree_view.set_cursor(path, self.vc_col, FALSE)
                tree_view.scroll_to_cell(path, self.vc_col, TRUE, 0.5, 0.0)
                tree_view.grab_focus ()
                break
            iter = self.videocard_store.iter_next(iter)
        

    def set_card_data(self, carddata, goto_row):
        name = carddata["NAME"];
        self.type_entry.set_text(name)
        self.driver_entry.set_text(carddata["DRIVER"])
        self.set_custom_memory(FALSE)
        if goto_row:
            self.goto_row(name)
        self.options_store.clear()
        if carddata.has_key("LINE"):
            lines = carddata["LINE"]
            self.parse_lines(lines)

    def probe_videocard(self, widget):
        primary = self.videocard.primaryCard()
        carddata = primary.getCardData()
        self.set_card_data(carddata, 1)

    def custom_mem_toggled(self, checkbutton):
        self.mem_optionmenu.set_sensitive(checkbutton.get_active())

    def set_custom_memory(self, custom):
        if custom:
            self.mem_check.set_active(TRUE)
            self.mem_optionmenu.set_sensitive(TRUE)
        else:
            self.mem_check.set_active(FALSE)
            self.mem_optionmenu.set_sensitive(FALSE)

    def set_memory_size(self, size):
        if size == 0:
            return
        
        ram_sizes = self.videocard.possible_ram_sizes()
        try:
            index = ram_sizes.index(size)
            self.mem_optionmenu.set_history(index)
        except:
            print _("Warning: Unknown videoram size %d!")%(size)

    def get_memory_size(self):
        if self.mem_check.get_active():
            ram_sizes = self.videocard.possible_ram_sizes()
            index = self.mem_optionmenu.get_history()
            return ram_sizes[index]
        return 0

    def selection_changed(self, selection):
        if self.changing_selection:
            return
        (store, iter) = selection.get_selected()
        if iter == None:
            return
        name = self.videocard_store.get_value(iter, 0)

        if self.videocard.cardsDB().has_key(name):
            carddata = self.videocard.cardsDB()[name]
            self.set_card_data(carddata, 0)
        
    def add_option(self, widget):
        iter = self.options_store.append()
        self.options_store.set_value(iter, 0, _("<Option Name>"))
        self.options_store.set_value(iter, 1, _("<Option Value, click to change>"))
        
    def remove_option(self, widget):
        selection = self.options_tree_view.get_selection()
        (store, iter) = selection.get_selected()
        if iter == None:
            return
        store.remove(iter)

    def option_name_edited(self, cell_renderer, path_string, text):
        iter = self.options_store.get_iter(string.atoi(path_string))
        self.options_store.set_value(iter, 0, text)
        
    def option_value_edited(self, cell_renderer, path_string, text):
        iter = self.options_store.get_iter(string.atoi(path_string))
        self.options_store.set_value(iter, 1, text)
        
        
    def hydrate(self, state):
        name = state.get_videocard_name()
        self.type_entry.set_text(name)
        self.goto_row(name)
        self.driver_entry.set_text(state.get_videocard_driver())
        self.set_memory_size(state.get_videocard_ram_or_probed())
        self.set_custom_memory(state.get_videocard_ram() != 0)
        self.options_store.clear()
        options = state.get_videocard_options()
        for option in options:
            iter = self.options_store.append()
            self.options_store.set_value(iter, 0, option[0])
            if option[1] == None:
                self.options_store.set_value(iter, 1, "")
            else:
                self.options_store.set_value(iter, 1, option[1])

    def dehydrate(self, state):
        state.set_videocard_name(self.type_entry.get_text())
        state.set_videocard_driver(self.driver_entry.get_text())
        state.set_videocard_ram(self.get_memory_size())

        #This doesnt work in the current pygtk if the list is empty. reported upstream
        #iter = self.options_store.get_iter_root()
        list = []
        iter = self.options_store.get_iter_first()
        while iter:
            list.append( (self.options_store.get_value(iter, 0), self.options_store.get_value(iter, 1)) )
            iter = self.options_store.iter_next(iter)
        state.set_videocard_options(list)
        
        
    def run(self):
        dialog = self.xml.get_widget("videocard_dialog")
        dialog.present()
        while 1:
            res = dialog.run()
            
            if res != gtk.RESPONSE_OK:
                dialog.hide()
                return FALSE
            
            break
        
        dialog.hide()
        return TRUE

    
class XConfigDialog:
    def setup_sizegroups(self):
        xml = self.xml
        size_group = gtk.SizeGroup(gtk.SIZE_GROUP_HORIZONTAL)
        size_group.add_widget(xml.get_widget("monitor_button"))
        size_group.add_widget(xml.get_widget("dpi_button"))
        size_group.add_widget(xml.get_widget("videocard_button"))
        
        size_group = gtk.SizeGroup(gtk.SIZE_GROUP_HORIZONTAL)
        size_group.add_widget(xml.get_widget("adv_monitor_table_label"))
        size_group.add_widget(xml.get_widget("adv_horiz_table_label"))
        size_group.add_widget(xml.get_widget("adv_dpi_table_label"))
        size_group.add_widget(xml.get_widget("adv_vert_table_label"))
        size_group.add_widget(xml.get_widget("adv_card_table_label"))
        size_group.add_widget(xml.get_widget("adv_card_mem_table_label"))
        size_group.add_widget(xml.get_widget("adv_card_driver_table_label"))

    def calc_dpi_string(self):
        (pw, ph) = self.state.get_physical_size_or_probed()
        res = self.state.get_resolution()
        return dpi_string(res, pw, ph)

    def update_ui(self):
        res = self.state.get_resolution()
        res_radio = self.resolution_radio[res]

        if not res_radio.get_active():
            res_radio.set_active(TRUE)
            
        (w,h) = resolution_from_string(res)
        self.size_preview.set_resolution(w,h)

        availible = self.state.availible_resolutions()
        for (res, res_radio) in self.resolution_radio.iteritems():
            if res in availible:
                res_radio.set_sensitive(TRUE)
            else:
                res_radio.set_sensitive(FALSE)

        depth = self.state.get_colordepth()
        depth_radio = self.xml.get_widget("radio_" + repr(depth) +"bit")

        if not depth_radio.get_active():
            depth_radio.set_active(TRUE)

        availible = self.state.availible_color_depths()
        for depth in all_depths:
            depth_radio = self.xml.get_widget("radio_" + repr(depth) +"bit")
            if depth in availible:
                depth_radio.set_sensitive(TRUE)
            else:
                depth_radio.set_sensitive(FALSE)


        label = self.xml.get_widget("display_label")
        label.set_text(self.state.get_monitor_name() + " with " + self.state.get_videocard_name())

        label = self.xml.get_widget("monitor_type_label")
        label.set_text(self.state.get_monitor_name())
        
        label = self.xml.get_widget("videocard_type_label")
        label.set_text(self.state.get_videocard_name())

        label = self.xml.get_widget("videocard_mem_label")
        mem = self.state.get_videocard_ram_or_probed()
        if mem == 0:
            label.set_text(_("Unknown"))
        elif mem %1024 != 0:
            label.set_text(_("%d kilobytes")%(mem))
        else:
            label.set_text(_("%d megabytes")%(mem/1024))

        label = self.xml.get_widget("videocard_driver_label")
        label.set_text(self.state.get_videocard_driver())

        checkbutton = self.xml.get_widget("videocard_dri_check")
        dri_supported = self.state.supports_dri()
        checkbutton.set_sensitive(dri_supported)
        checkbutton.set_active(dri_supported and self.state.get_dri_enabled())

        label = self.xml.get_widget("monitor_horiz_label")
        label.set_text(self.state.get_hsync_or_probed())
        
        label = self.xml.get_widget("monitor_vert_label")
        label.set_text(self.state.get_vsync_or_probed())
            
        label = self.xml.get_widget("monitor_dpi_label")
        label.set_text(self.calc_dpi_string())


    def resolution_toggled(self, radiobutton):
        if radiobutton.get_active():
            res = radiobutton.get_data("resolution")
            self.state.set_resolution(res)
            self.update_ui()
            
    def colordepth_toggled(self, radiobutton):
        if radiobutton.get_active():
            depth = radiobutton.get_data("colordepth")
            self.state.set_colordepth(depth)
            self.update_ui()
            
    def dri_toggled(self, checkbutton):
        self.state.set_dri_enabled(checkbutton.get_active())
        self.update_ui()
        
    def setup_ui_resolution(self):
        box = self.xml.get_widget("resolution_radiobutton_box")
        last = None
        all_resolutions = self.state.get_all_resolutions()
        for res in all_resolutions:
            widget = gtk.RadioButton(last, res)
            last = widget
            box.pack_start(widget, FALSE, FALSE, 0)
            widget.show()
            widget.connect("toggled", self.resolution_toggled)
            widget.set_data("resolution", res)
            self.resolution_radio[res] = widget
        
    def setup_ui_colordepth(self):
        for depth in all_depths:
            widget = self.xml.get_widget("radio_"+repr(depth) + "bit")
            widget.connect("toggled", self.colordepth_toggled)
            widget.set_data("colordepth", depth)
            
    def run(self):
        if xserverpid:
            xpos = gtk.gdk.screen_width() / 2 - \
                   self.xml.get_widget("display_dialog").size_request()[0] / 2
            ypos = gtk.gdk.screen_height() / 2 - \
                   self.xml.get_widget("display_dialog").size_request()[1] / 2
            self.xml.get_widget("display_dialog").set_uposition(xpos, ypos)
        self.xml.get_widget("display_dialog").show()
        self.dialog_result = 0
        gtk.main()
        self.xml.get_widget("display_dialog").hide()
        return self.dialog_result

    def ok(self, widget):
        self.dialog_result = 1
        gtk.main_quit()

    def cancel(self, widget):
        gtk.main_quit()
        
    def delete(self, widget, event):
        gtk.main_quit()
        return 1

    def configure_monitor(self, widget):
        if not self.monitor_dialog:
            self.monitor_dialog = MonitorDialog(self.xml, self.state.monitor,
                                                self.state.monitor.getMonitorID(TRUE))
            
        self.monitor_dialog.hydrate(self.state)
        if self.monitor_dialog.run():
            self.monitor_dialog.dehydrate(self.state)
            self.state.recalc_mode()
            self.update_ui()
        
    def configure_videocard(self, widget):
        if not self.videocard_dialog:
            self.videocard_dialog = VideocardDialog(self.xml, self.state.videocard)
            
        self.videocard_dialog.hydrate(self.state)
        if self.videocard_dialog.run():
            self.videocard_dialog.dehydrate(self.state)
            self.state.recalc_mode()
            self.update_ui()
        
    def configure_dpi(self, widget):
        if not self.dpi_dialog:
            self.dpi_dialog = DPIDialog(self.xml, self.state.monitor)
            
        self.dpi_dialog.hydrate(self.state)
        if self.dpi_dialog.run():
            self.dpi_dialog.dehydrate(self.state)
            self.update_ui()
        
    def __init__(self, hardware_state):
        # setup UI
        gtk.glade.bindtextdomain("redhat-config-xfree86")
        if os.access("display.glade", os.F_OK):
            xml = gtk.glade.XML("display.glade", domain="redhat-config-xfree86")
        else:
            xml = gtk.glade.XML("/usr/share/redhat-config-xfree86/ui/display.glade", domain="redhat-config-xfree86")
            
        self.state = hardware_state
        self.xml = xml
        self.size_preview = ScreenSizePreview(xml.get_widget("screenshot_darea"),200,150)

        self.resolution_radio = {}
        self.videocard_dialog = None
        self.monitor_dialog = None
        self.dpi_dialog = None

        self.setup_sizegroups()
        self.setup_ui_resolution()
        self.setup_ui_colordepth()
        
        xml.get_widget("videocard_dri_check").connect("toggled", self.dri_toggled)
        xml.get_widget("monitor_button").connect("clicked", self.configure_monitor)
        xml.get_widget("videocard_button").connect("clicked", self.configure_videocard)
        xml.get_widget("dpi_button").connect("clicked", self.configure_dpi)

        xml.get_widget("ok_button").connect("clicked", self.ok)
        xml.get_widget("cancel_button").connect("clicked", self.cancel)
        xml.get_widget("display_dialog").connect("delete_event", self.delete)

def generate_xconfig(hardware_state):
    xconfig = xf86config.createTemplate()

    hardware_state.merge_into(xconfig)
    
    keyboard = rhpl.keyboard.Keyboard()
    keyboard.read()

    def get_option(device, optionname):
        res = filter (lambda o: o.name == optionname, device.options)
        if len (res) > 0:
            return res[0]
        else:
            option = xf86config.XF86Option(optionname)
            device.options.insert(option)
            return option

    corekb = xf86config.getCoreKeyboard(xconfig)
    
    get_option(corekb, "XkbRules").val = keyboard["rules"]
    get_option(corekb, "XkbModel").val = keyboard["model"]
    get_option(corekb, "XkbLayout").val = keyboard["layout"]
    if keyboard["options"] != "":
        get_option(corekb, "XkbOptions").val = keyboard["options"]
    if keyboard["variant"] != "":
        get_option(corekb, "XkbVariant").val = keyboard["variant"]
       
# FIXME: I have no idea why, but on my machine i had to probe twice to
# find the mouse
    mouse = rhpl.mouse.Mouse()
    mouse = rhpl.mouse.Mouse()

    corepointer = xf86config.getCorePointer(xconfig)
    type = mouse.info["XMOUSETYPE"]
    if type == "none":
        type = "PS/2" # No idea, Guess PS/2
    get_option(corepointer, "Protocol").val = type
    get_option(corepointer, "Device").val = "/dev/" + mouse.getDevice()
    (typename, emu3) = mouse.get()
    val = "no"
    if emu3:
        val = "yes"
    get_option(corepointer, "Emulate3Buttons").val = val

    if mouse.getDevice() == "input/mice":
        inputs = xconfig.layout[0].inputs
        for i in range(len(inputs)):
            if inputs[i].inputdev == "DevInputMice":
                inputs.remove(i)
                break
    
    return xconfig

def finished_dialog(config_file, backup_file, moved_old_file):
    message = _("Display settings changed\n\n")
    if not xserverpid:
        message = message + _("You need to log out and restart the X server for the changes to take effect.\n")
    if backup_file:
        message = message + _("Configuration was written to %s, original configuration saved as %s.\n") % (config_file, backup_file)
    else:
        message = message + _("Configuration was written to %s.\n") % (config_file)
    if moved_old_file:
        message = message + _("\nNote: The file /etc/X11/XF86Config-4 has been deprecated and we now always use /etc/X11/XF86Config. The XF86Config-4 file has been saved as XF86Config-4.deprecated.")
        
    m = gtk.MessageDialog(None, gtk.DIALOG_MODAL,
                          gtk.MESSAGE_INFO, gtk.BUTTONS_OK, message)
    m.run()
    m.destroy()

def fork_off_server(xconfig):
    tmp_config = tempfile.mktemp("xf86config")
    verbose(_("Writing temporary config file to %s") % tmp_config)
    xconfig.write(tmp_config)
    os.environ["DISPLAY"]=":17.0"
    verbose(_("Trying to start X server"))
    args = [ "/usr/X11R6/bin/XFree86", "-xf86config", tmp_config, "-ac", "-v", "-nolisten", "tcp", "-s", "1440", "-terminate",
             "-dpms", "-logfile", "/dev/null", ":17"]
    
    serverpid = os.fork()
    if serverpid == 0: #child
        try:
            err = os.open(logFile, os.O_RDWR | os.O_CREAT, 0644)
            if err < 0:
                sys.stderr.write(_("error opening %s\n" % logFile))
            else:
                os.dup2(err, 2)
                os.close(err)
        except:
            # oh well
            pass
        
        os.execv(args[0], args)
        sys.exit (1)
            
    return (serverpid, tmp_config)

def start_x_server(xconfig):
    try:
        (server, tmp_config) = fork_off_server(xconfig)
    except:
	import traceback
        server = None
	(type, value, tb) = sys.exc_info()
	list = traceback.format_exception (type, value, tb)
	text = string.joinfields (list, "")
	print text
    
    # give time for the server to fail (if it is going to fail...)
    if not server:
        verbose(_("X SERVER FAILED"));
        raise RuntimeError, "X server failed to start"

    count = 0

    if verbose_output:
        sys.stdout.write(_("Waiting for X server to start...log located in %s\n") % logFile)
        sys.stdout.flush()

    for i in range(5):
        time.sleep(1)
        if verbose_output:
            sys.stdout.write("%s..." % (i+1))
            sys.stdout.flush()
        
    while count < 60:
        if verbose_output:
            sys.stdout.write(".")
            sys.stdout.flush()
        pid = 0
        try:
            pid, status = os.waitpid (server, os.WNOHANG)
        except OSError, (errno, msg):
            print __name__, "waitpid:", msg
	if pid:
            verbose(_("X SERVER FAILED"));
	    raise RuntimeError, "X server failed to start"
	try:
	    os.stat ("/tmp/.X11-unix/X17")
	    break
	except OSError:
	    pass
	time.sleep(1)
	count = count + 1
  
    # We dont need the configuration file anymore.
    os.unlink(tmp_config)

    if count < 60:
        verbose (_("X server started successfully."))
        return server
    else:
        return 0

def handle_set_commands(set_values, hardware_state):
    # Set up values that limit resolution first
    if set_values.has_key("hsync"):
        hardware_state.set_hsync(set_values["hsync"])
    if set_values.has_key("vsync"):
        hardware_state.set_vsync(set_values["vsync"])
    if set_values.has_key("videoram"):
        hardware_state.set_videocard_ram(string.atoi(set_values["videoram"]))
    if set_values.has_key("driver"):
        hardware_state.set_videocard_driver(set_values["driver"])
    if set_values.has_key("card"):
	hardware_state.set_videocard_card(set_values["card"])
    if set_values.has_key("resolution"):
        hardware_state.set_resolution(set_values["resolution"])
    if set_values.has_key("depth"):
        depth = string.atoi(set_values["depth"])
        hardware_state.set_colordepth(depth)

def kill_x_server():
    if xserverpid:
        try:
            os.kill(xserverpid, 15)
            os.waitpid(xserverpid, 0)
        except:
            pass
   

def usage():
    print \
_("""Usage: redhat-config-xfree86 [OPTIONS]
Options:
  -h, --help        display this help and exit
  -v, --verbose     display what the program is doing more verbosely
  -o, --output=     the filename of the config file to be output
      --reconfig    don't base configuration on existing config files
      --noui        don't show the gui, implicit when using --set options
      --forceui     force the ui to be shown, needed when using --set options
                    to change some setting, but you still want to show the ui
      --set-<key>=  change the value of a specific configuration key.

        currently supported keys are:

        resolution  the screen resolution used
        depth       the color depth in bits
                    most drivers support 8, 15, 16 and 24
        driver      the graphics card driver to use
        card        the graphics card description to use
                    e.g. 'ATI Radeon (generic)'
        vsync       monitor vertical sync rates allowed (in Hz)
        hsync       monitor horizontal sync rates allowed (in kHz)
        videoram    the amount of videoram in kilobytes, 0 means probe
                    
""")

if __name__ == '__main__':
    signal.signal(signal.SIGINT, signal.SIG_DFL)

textdomain("redhat-config-xfree86")

output_file = "/etc/X11/XF86Config"
output_file_specified = FALSE

try:
    (opts, rest_args) = getopt.getopt(sys.argv[1:],
                               "vo:h",
                               ["help", "reconfig", "noui", "verbose",
                                "forceui", "output=",
                                "set-hsync=", "set-vsync=",
                                "set-videoram=", "set-driver=", "set-card="
                                "set-resolution=", "set-depth="])
except (getopt.GetoptError), e:
    print e
    print
    usage()
    sys.exit(1)

reconfig = FALSE
no_ui = FALSE
force_ui = FALSE
set_values = {}
verbose_output = FALSE

for (opt, value) in opts:
    if opt == "-v" or opt == "--verbose":
        verbose_output = TRUE
    if opt == "--reconfig":
        reconfig = TRUE
    if opt == "-h" or opt == "--help":
        usage()
        sys.exit(1)
    if opt == "--noui":
        no_ui = TRUE
    if opt == "--forceui":
        force_ui = TRUE
    if opt == "-o" or opt == "--output":
        output_file = value
        output_file_specified = TRUE
    if opt[:6] == "--set-":
        no_ui = TRUE
        set_values[opt[6:]] = value

if force_ui:
    no_ui = FALSE

for loop_count in [1,2]:
    xconfig = None
    xconfigpath = None
    xserverpid = 0

    if not reconfig:
        (xconfig, xconfigpath) = xf86config.readConfigFile()
        if xconfig:
            verbose(_("Read configuration file %s") % xconfigpath)
        else:
            verbose(_("Could not find existing X configuration"))

    hardware_state = HardwareState(xconfig)

    if xconfig == None:
        xconfig = generate_xconfig(hardware_state)

    handle_set_commands(set_values, hardware_state)

    if not no_ui:
        no_x = FALSE
        try:
            import gtk
            reload(gtk)
        except:
            no_x = TRUE

        if no_x:
            # X not already running.  Try to bring up a server.
            try:
                xserverpid = start_x_server(xconfig)
            except:
                reconfig = TRUE
                xconfig = None
                if loop_count == 1:
                    print _("Couldn't start X server, trying with a fresh configuration")
                    continue
                else:
                    print _("Error, cannot start X server.")
                    sys.exit(1)

        if xserverpid:
            # if we had to start an X server, make it look half-way decent.
            import gtk
            reload(gtk)        
            import gtk.gdk
            root = gtk.gdk.get_default_root_window()
            arrow = gtk.gdk.Cursor(gtk.gdk.LEFT_PTR)
            root.set_cursor (arrow)

        # Work around bug in pygtk where it sets LC_NUMERIC
        locale.setlocale(locale.LC_NUMERIC,"C")
        import gtk.glade
        # TODO: try to import gtk+, on fail, try to start new X
        dialog = XConfigDialog(hardware_state)
        dialog.update_ui()
        if not dialog.run():
            kill_x_server()
            sys.exit(1)
    break

hardware_state.merge_into(xconfig)

if xconfigpath and not output_file_specified:
    output_file = xconfigpath
    if output_file == "/etc/X11/XF86Config-4":
        output_file = "/etc/X11/XF86Config"

moved_old_file = FALSE
if os.access("/etc/X11/XF86Config-4", os.F_OK):
    verbose(_("Moving /etc/X11/XF86Config-4 to /etc/X11/XF86Config-4.deprecated"))
    os.rename("/etc/X11/XF86Config-4", "/etc/X11/XF86Config-4.deprecated")
    moved_old_file = TRUE
            
backup_file = None
if os.access(output_file, os.F_OK):
    backup_file = output_file + ".backup"
    verbose(_("Backing up %s to %s") % (output_file, backup_file))
    os.rename(output_file, backup_file)

verbose(_("Writing configuration to %s") % output_file)
xconfig.write(output_file)

if os.access("/etc/X11/X", os.F_OK):
    verbose(_("Removing old /etc/X11/X"))
    os.unlink("/etc/X11/X")

verbose(_("Creating /etc/X11/X symlink"))
os.symlink("../../usr/X11R6/bin/XFree86", "/etc/X11/X")


# Tell gdm to restart the server when the user logs out
# Otherwise the xserver won't re-read the config file
gdmfifofile = "/var/gdm/.gdmfifo"
if os.access(gdmfifofile, os.F_OK):
    verbose(_("Kicking gdm"))
    try:
        fd = os.open(gdmfifofile, os.O_WRONLY | os.O_APPEND | os.O_NONBLOCK)
        if fd >= 0:
            os.write(fd,"\nDIRTY_SERVERS\n");
            os.close(fd)
    except:
        pass

if not no_ui:
    finished_dialog(output_file, backup_file, moved_old_file)

kill_x_server()

